#include <stdio.h>                      // printf()
#include "../intel/timestmp.h"          // PentiumClock
#include "../apptel/rxfpga.h"           // RxFpga_registers
#include "../apptel/sram.h"             // prototypes we must adhere to



// manifest constants
//
#define MAX_CYCLE_DELAY     1.0         // in seconds



// global variables for statistics
//
static PentiumClock timing_min, timing_max, timing_tot;
static Bit32 iterations_min, iterations_max, iterations_tot;
static Bit32 num_calls;



// initialize statistics
//
void sram_init_stats( void )
{
    num_calls = 0;

    timing_min.low = timing_min.high = BIT32_MAX;
    timing_max = (Bit32) 0;
    timing_tot = (Bit32) 0;

    iterations_min = BIT32_MAX;
    iterations_max = 0;
    iterations_tot = 0;
}



void sram_print_stats( void )
{
    if( !num_calls )
    {
        fprintf( stdout, "sram_read_or_write() never called successfully\n" );
        return;
    }

    fprintf( stdout, "sram_read_or_write() called %u times successfully\n", num_calls );

    fprintf( stdout, "timing: min/max/mean %g/%g/%g uS\n",
        1e6 * timing_min.get_seconds(),
        1e6 * timing_max.get_seconds(),
        1e6 * timing_tot.get_seconds() / num_calls
    );

    fprintf( stdout, "iterations: min/max/mean %u/%u/%g\n",
        iterations_min, iterations_max,
        (double)iterations_tot / num_calls
    );
}



extern "C"
int sram_read_or_write( RxFpga_registers *rx_fpga, Bit32 address, int want_write, Bit32 *data0 )
{
    PentiumClock now, alarm, before;
    Bit32 iterations;
    int cycle_is_pending;
    RxSramAddress2Control_register address2control;

    if( !rx_fpga )
        return ERROR_SRAM_NOT_INITIALIZED;

    if( address > MAX_SRAM_ADDRESS )
    {
        fprintf( stderr, "address 0x%.8X is larger than maximum 0x%.8X\n", address, MAX_SRAM_ADDRESS );
        return ERROR_SRAM_INVALID_ADDRESS;
    }

    // a word to the wise: if you do these register writes
    // in order of ascending addresses, the host locks up hard,
    // requiring a hard reset or power-cycle
    //
    // here's why: (really a CPLD s/w bug)
    // doing register writes in order causes the Intel
    // 83439HX PCI controller on the motherboard to generate
    // a burst on the PCI bus, and during the 2nd data phase
    // the AMCC PCI i/f on the POINT card determines that
    // a burst is in progress, and it leaves PTATN# asserted,
    // but the CPLD on the board assumes that signal will
    // always be deasserted before reassertion (single cycles),
    // so it never asserts the PTRDY# signal for the 3rd+ data
    // phases, and hence the AMCC will issue a PCI retry request,
    // but since the CPLD never sees a PTATN# deassertion, when
    // the host tries again 272 nS later it will fail again,
    // ad infinitum, ad nauseum
    //

    if( want_write )
    {
        Bit32 data = *data0;

        // split the 32 bit data value up into 4 bytes
        // and put it directly into the RX FPGA registers
        //
        rx_fpga->SramData3.Value = (data >> 24) & 255;
        rx_fpga->SramData2.Value = (data >> 16) & 255;
        rx_fpga->SramData1.Value = (data >> 8) & 255;
        rx_fpga->SramData0.Value = data & 255;
    }

    // put the low 2 bytes of the address directly into the
    // RX FPGA registers
    //
    rx_fpga->SramAddress1.Value = (address >> 8) & 255;
    rx_fpga->SramAddress0.Value = address & 255;

    // put the high 4 bits of the address into a host-based
    // temporary area, so we can write it, along with the
    // command, all at once
    //
    address2control.Value = (address >> 16) & 15;

    // fill in the control fields of same register
    //
    address2control.Reserved = 0;
    address2control.DoCycle = TRUE;
    address2control.CycleIsWrite = want_write;

    // now copy the high bits of address and control to RX FPGA register
    //
    rx_fpga->SramAddress2Control = address2control;

    // wait for the cycle to finish or a timeout
    //
    alarm.put_seconds( MAX_CYCLE_DELAY );
    before.set();
    alarm += before;
    iterations = 0;
    do
    {
        cycle_is_pending = rx_fpga->SramAddress2Control.DoCycle;
        now.set();
        iterations++;

    } while( cycle_is_pending && now<=alarm );

    if( cycle_is_pending )
    {
        fprintf( stderr, "error: more than %f seconds (%u iterations) elapsed"
            " waiting for SRAM cycle to complete\n", MAX_CYCLE_DELAY, iterations );
        return ERROR_SRAM_CYCLE_TIMEOUT;
    }

    if( !want_write )

        // recompose the word from the 4 byte register
        //
        *data0 = rx_fpga->SramData0.Value
              | (rx_fpga->SramData1.Value << 8)
              | (rx_fpga->SramData2.Value << 16)
              | (rx_fpga->SramData3.Value << 24);

    // update our stats
    //
    num_calls++;

    iterations_tot += iterations;
    if( iterations_max < iterations )
        iterations_max = iterations;
    if( iterations_min > iterations )
        iterations_min = iterations;

    now -= before;
    timing_tot += now;
    if( timing_max < now )
        timing_max = now;
    if( timing_min > now )
        timing_min = now;

    return 0;
}



