#include <stdio.h>                      // printf()
#include <string.h>                     // strstr()
#include <mem.h>                        // memset()
#include <errno.h>                      // extern errno

#include "../dpmi32/phys.h"
#include "../dpmi32/intr.h"
#include "../pci/pci.h"
#include "../integrat/constant.h"       // CARD_CLOCK_HEADER CARD_CLOCK_DRIVER
#include CARD_CLOCK_HEADER              // class CardClock
#include CARD_DRIVER_HEADER             // prototypes we must adhere to
#include "../intel/byteswap.h"          // ntohl()
#include "../integrat/trace.h"          // struct tracerec



// conditional compilation switches
//
#define OLD_AALI                        FALSE
#define UART_TALKS                      TRUE
#define UART_LISTENS                    FALSE       // TRUE to type commands to ATM cards' ROM monitor (but interferes with flows_keystroke)
#define ALLOW_SONET_INTERRUPTS          FALSE
#define ALLOW_ATM_INTERRUPTS            TRUE
#define ALLOW_ATM_RX_FREE_EMPTY_IRQ     TRUE        // FALSE to only collect 2 blocks
#define ONE_CARD_PER_TIMER_IRQ          TRUE
#define ESSENTIAL_STATUS_QUERIES        FALSE       // TRUE does 2 extra PCI slave reads (slows card)
#define UNEXPECTED_ATM_IRQ_PARANOIA     FALSE       // TRUE does 1 extra PCI slave read  (slows card)
#define DEBUG_STACK_OVERFLOW            FALSE
#define ATM_EXIT_ITERATES_TIMER_HANDLER FALSE       // TRUE to test w/o interrupts
#define ATM_EXIT_FILLS_PARTIAL_BLOCK    TRUE        // FALSE to see only what interrupt handlers gave you
#define ATM_EXIT_RELEASES_BLOCKS        TRUE
#define ATM_EXIT_DUMPS_ATM_HEADERS      FALSE
#define ATM_EXIT_DUMPS_TIMES            FALSE
#define PENDING_BLOCK_PARANOIA          TRUE
#define ATM_INIT_RESETS_SAR             TRUE
#define ATM_INIT_SETS_PCI_TIMING        FALSE       // may contribute to lockup on 82439HX with 200 MHz CPU when TRUE
#define ATM_INIT_SETS_PCI_CONTROL       FALSE       // causes host CPU to get too few PCI bus cycles
#define ATM_INIT_SETS_CACHING           FALSE
#define ATM_SHIFT_CLEANUP_DEBUG         FALSE
#define SAVE_INTERRUPT_THIS_PTR         FALSE
#define SWAP_FIRMWARE                   FALSE
#define COPY_FIRMWARE_HEADER_TO_CARD    TRUE
#define CHECK_MEMORY_WHILE_EXITING      FALSE       // extra checks while >=1 card still running, affects RX free ring unavail count
#define DISABLE_TIMER_REENTRY           FALSE       // thought it would allow debugger to single-step interrupt handler, but it doesn't
#define CARD_KEEPS_FIFO_STATS           TRUE
#define SHOW_FIFO_DEPTHS                FALSE



// compile-time switches which depend on others
//
#define PCI_EXIT_CALLS_BUFFER_DUMP      !DO_FOREGROUND_STATS

#if (CELLS_PER_BUFFER > 0) && !CARD_COPIES_ALL_CELLS
#define CARD_HANDLES_TIMESTAMP_WRAP     TRUE
#else
#define CARD_HANDLES_TIMESTAMP_WRAP     FALSE
#endif



// manifest constants
//
#define ATM_CALIBRATE       5           // record interrupt timestamp here

#define SONET_RESET_DELAY   60          // 1/2 microsecond worth of 120 MHz Pentium clocks
#define ATM_RESET_DELAY     64.0        // count of PCI clocks
#define ROM_START_DELAY     15.0        // count of (floating point) seconds
#define FIRM_START_DELAY    5.0         // count of (floating point) seconds
#define FIRM_INIT_DELAY     1.0         // count of seconds
#define SHUTDOWN_DELAY     10.0         // count of seconds (10 S)
//#define BEAT_CHANGE_DELAY 100.0         // count of seconds (100 S)
#define BEAT_CHANGE_DELAY  10.0         // count of seconds (10 S)
#define BEAT_CHANGE_HOLDOFF 0.1         // count of seconds
//#define TIME_CLEAR_DELAY    0.001       // count of seconds (1 mS)
#define TIME_CLEAR_DELAY  100.0         // count of seconds (100 S)
#define UART_DELAY_COMMAND  0.1         // count of seconds
#define UART_DELAY_RUNNING  0.0         // count of seconds
#define UART_DELAY_INITIAL  0.1         // count of seconds
#define UART_DELAY_HEART    0.1         // count of seconds
#define UART_DELAY_FINAL    0.5         // count of seconds
#define UART_DELAY_FAILURE  3600.0      // count of seconds (1 hour)
#define UART_DELAY_SHUTDOWN 1.0         // count of seconds (1 S)

#define DESIRED_MIN_GRANT   2

//#define DESIRED_CARD_LATENCY        0   // only setting that allows 2 cards to keep up with IP-header-only capture
#define DESIRED_CARD_LATENCY      8   // raw capture uses 10 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     16   // raw capture uses 7 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     32   // raw capture uses 6 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY     64   // (default) raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY    128   // raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells
//#define DESIRED_CARD_LATENCY    248   // raw capture uses 5 RX FIFO slots with buf=4, 48-byte cells

#define UART_BUFFER_SIZE         1000   // how much input do we care about?

//#define HEARTBEAT_CHECKS          3   // how many times do we check heartbeat to be sure it is increasing/decreasing?
//#define HEARTBEAT_CHECKS          5   // how many times do we check heartbeat to be sure it is increasing/decreasing?
#define HEARTBEAT_CHECKS            0   // how many times do we check heartbeat to be sure it is increasing/decreasing?

#define MAX_FIFO_DEPTH           2047   // how many buckets for possible SONET-to-i960 FIFO depths?





#if SAVE_INTERRUPT_THIS_PTR
    static class AtmCardRunning *atm_this_pointer_during_interrupt=(AtmCardRunning *)0x12345678;         // debug only!
#endif
#if DEBUG_STACK_OVERFLOW
    static int stack_ptr_recorded = FALSE;
    static Bit16 stack_seg_during_class_timer = 0;
    static Bit32 stack_ptr_during_class_timer = 0;
#endif



// must declare class static variables
// so they get allocated and initialized
//
InterruptController AtmCardRunning::irq;



// re-route the card interrupt to our code
//
void AtmCardRunning::interrupt_init( void )
{
    irq_level = pci_get_irq_level( card_handle );
    interrupt_number = irq_to_intr( irq_level );
    old_handler = getvect( interrupt_number );
    disable();

#if ALLOW_ATM_INTERRUPTS || ALLOW_SONET_INTERRUPTS
    thunk = make_thunk( (ifp)interrupt_handler0, &old_handler, this );
#if THROW_IS_LEGAL
    if( thunk == NULL )
        throw "error!  cannot allocate card interrupt thunk!\n";
#endif

    Bit48 pointer;
    pointer.offset = (Bit32) thunk;
    pointer.selector = _CS;
    setvect( interrupt_number, pointer );

    irq.unmask( irq_level );
#else
    irq.mask( irq_level );
#endif
    enable();
}



void AtmCardRunning::interrupt_exit( void )
{
    disable();
#if ALLOW_ATM_INTERRUPTS || ALLOW_SONET_INTERRUPTS
    setvect( interrupt_number, old_handler );
    free( thunk );
#endif
    enable();
}



// deals with interrupts from any of 2 sources
// (SONET framer chip and/or ATM SAR chip) and 2 cards
// (monitoring send and receive fibers for a single link)
//
// since we poll the shared memory, we only use
// this interrupt to report severe errors
//
int AtmCardRunning::interrupt_handler( void )
{
#if ESSENTIAL_TIME_STAMPS
#endif

    // assume ATM SAR wants attention...
    //
    atm_handler();

    // tell the interrupt controllers we are done
    //
    irq.acknowledge( irq_level );

#if ESSENTIAL_TIME_STAMPS
#endif

    // tell the thunking code not to chain to the old handler
    //
    return 0;
}



// initialize the SONET framer chip
//
void AtmCardRunning::sonet_init( void )
{
    // we haven't had any interrupts from
    // this chip on this card yet
    //
    sonet_interrupt_count = 0;
}



// deals with interrupts from the TI TNETA1500 SONET framer
//
// (alarms and MIB counters not implemented yet)
//
void AtmCardRunning::sonet_handler( void )
{
    // set a flag to indicate that this occurred
    //
    sonet_interrupt_count++;

    // maybe retrieve the counters?
    //
}



void AtmCardRunning::sonet_exit( void )
{
    // dump our meager statistics
    //
    printf( "card 0x%4.4X had %ld SONET framer interrupts\n", card_handle, sonet_interrupt_count );
}



// writes any output you give me to the software UART,
// reads any input already available from the UART,
// and waits the given amount of time for more input
//
// based on conditional compile symbols, this code may
// also display input from UART on standard output and/or
// accept keystrokes to output to the card
//
// prefixes each line displayed on STDOUT so you can tell
// which card provided it, and that it came from the UART
//
// assumes input_size is nonzero when input is non-NULL
//
// will always terminate the input string
//
void AtmCardRunning::atm_uart_io( const char *output, char *input, unsigned input_size, double delay )
{
    char c;
    Soft_UART *uart = &memmap->shared.mon960.uart;
    PentiumClock now, initial, alarm;
    double timing;
#if UART_TALKS
    char signon[ 100 ];
#endif

#if UART_TALKS
    sprintf( signon, "card 0x%4.4X UART: ", card_handle );
#endif

    // preallocate the terminating NUL character
    //
    input_size--;

    initial.set();
    alarm.put_seconds( delay );
    alarm += initial;

    // at least once...
    //
    do
    {
        now.set();

        // if there is input ready for us...
        //
        if( CAN_GETC_i960( uart ) )
        {
            // read it into a local variable
            //
            GETC_i960( uart, c );

#if UART_TALKS
            // if nothing has been output on this screen line
            // and this character is not going to wrap back to
            // the beginning of the line...
            //
            if( atm_uart_new_line && c!='\r')
            {
                // remember we output on this line
                //
                atm_uart_new_line = FALSE;

                // insert a prefix message before the character
                //
                fputs( signon, stdout );
            }

            // show on screen
            //
            putchar( c );

            // if this was a newline character...
            //
            if( c == '\n' )

                // forget we have output any lines
                //
                atm_uart_new_line = TRUE;
#endif

            // if the caller wants input,
            // and there is room left...
            //
            if( input && input_size )
            {
                // save input character for caller
                //
                *input++ = c;

                // less space available now
                //
                input_size--;
            }

            // reset timer
            //
            initial = now;
            alarm.put_seconds( delay );
            alarm += initial;
        }

        // if there is room in output buffer...
        //
        if( CAN_PUTC_i960(uart,*output) )
        {
            // if the caller wants to output
            // and we haven't done all of that yet...
            //
            if( output && *output )
            {
                c = *output++;
            }
#if UART_LISTENS
            else if( kbhit() )
            {
                c = getch();
                if( c == 27 )           // escape
                    break;
            }
#endif
            else
                c = 0;

            if( c )
            {
                // spit out one character
                //
                PUTC_i960( uart, c );

                // reset timer
                //
                initial = now;
                alarm.put_seconds( delay );
                alarm += initial;
            }
        }
    }
    // ...as long as time hasn't run out
    //
    while( now < alarm );


    // terminate the input string
    //
    if( input )
        *input = '\0';
}



// called from the foreground when it is "idle"
// (not very often)
//
// we use this opportunity to service the soft-UARTs,
// in case the card has any diagnostic output for us
//
void AtmCardRunning::tick( void )
{
    atm_uart_io( NULL, NULL, 0, UART_DELAY_RUNNING );
}



int AtmCardRunning::atm_reset_hardware( void )
{
    // set the boot status word to cold start, so the
    // i960 ?ROM code will do the right thing after we reset it
    //
    memmap->shared.mon960.bstat = cold_start;

    union { HostControl_write hcr_w; Bit32 hcr_w_alias; };

    // save any PCI registers that will get trashed by the reset
    //
    // ???apparently none get trashed
    //

    // start a software reset
    //
    // FORE recommends 8-bit accesses for this,
    // although testing shows that 32-bit works too
    //
    hcr_w_alias = 0;
    hcr_w.board_reset = TRUE;
#if 1
    memmap->dword_00h = hcr_w_alias;
#else
    memmap->hcr_w = hcr_w;
#endif

    // wait for the NIC to finish resetting
    //
    // FORE recommends 63 PCI clocks
    //
    PentiumClock::wait( (ATM_RESET_DELAY / PCI_CLOCK_RATE) * CLOCK_RATE );

    // stop resetting the card
    //
    // FORE recommends 8-bit accesses for this,
    // although testing shows that 32-bit works too
    //
    hcr_w.board_reset = FALSE;
#if 1
    memmap->dword_00h = hcr_w_alias;
#else
    memmap->hcr_w = hcr_w;
#endif

    // restore any PCI registers that got trashed by the reset
    //
    // ???apparently none were trashed
    //

    // wait a few seconds for the boot status word
    // to make a transition
    //
    Boot_status prev_status = cold_start;
    Boot_status status;
    PentiumClock now, initial, alarm;
    double timing;

    initial.set();
    now = initial;
    alarm.put_seconds( ROM_START_DELAY );
    alarm += now;

    while( now<alarm )
    {
        status = memmap->shared.mon960.bstat;
        if( status != prev_status )
        {
            now -= initial;
            timing = now.get_seconds();

            // check that self test succeeded
            //
            if( status == self_test_ok )
                break;

            if( status == self_test_fail )
            {
                printf( "card 0x%4.4X firmware self test failed after %g seconds\n", card_handle, timing );
                return ERROR_CARD_SELF_TEST_FAILED;
            }

            printf( "card 0x%4.4X warning: firmware transitioned to unknown status 0x%8.8lX after %g seconds\n", card_handle, status, timing );
            prev_status = status;
        }

        now.set();
    }

    if( status == prev_status )
    {
        printf( "card 0x%4.4X ROM stayed in 0x%8.8lX state past %g seconds\n", card_handle, status, ROM_START_DELAY );
        if( status == cold_start )
            return ERROR_CARD_BSTAT_STUCK;

        return ERROR_CARD_SELF_TEST_UNKNOWN;
    }

    // tell the user how long it took to succeed
    //
    printf( "card 0x%4.4X ROM took %g S to cold start and self test\n", card_handle, timing );

    return 0;
}



// download i960 code image to card
//
int AtmCardRunning::atm_download( void )
{
    FILE *handle;
    FirmwareHeader header;
    int stdio_result;
    Bit8 *start;
    Bit32 *current;
    Bit32 one_word;
    long size, count;

    // disable cr-lf to lf translation!
    //
    handle = fopen( FIRMWARE_FILE_NAME, "rb" );
    if( handle == NULL )
    {
        printf( "card 0x%4.4X error %d opening file '%s'\n", card_handle,
            errno, FIRMWARE_FILE_NAME );
        return ERROR_CARD_CANNOT_OPEN_FILE;
    }

    // get the file size
    //
    stdio_result = fseek( handle, 0, SEEK_END );
    if( stdio_result != 0 )
    {
        printf( "card 0x%4.4X error %d seeking to end of file '%s'\n",
            card_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_SEEK_END;
    }

    size = ftell( handle );
    if( size == -1L )
    {
        printf( "card 0x%4.4X error %d getting file '%s' position\n",
            card_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_TELL_SIZE;
    }

    // do we need to skip to a certain position in the file
    // to read the header???  it seems to be at 0x44 not 0...
    //
    fseek( handle, 0, SEEK_SET );
    if( stdio_result != 0 )
    {
        printf( "card 0x%4.4X error %d seeking to beginning of file '%s'\n",
            card_handle, errno, FIRMWARE_FILE_NAME );

        fclose( handle );
        return ERROR_CARD_CANNOT_SEEK_BEG;
    }

    // read in the file header
    //
    stdio_result = fread( &header, sizeof(FirmwareHeader), 1, handle );
    if( stdio_result != 1 )
    {
        printf( "card 0x%4.4X error %d reading file '%s' %d byte header; got %d items\n",
            card_handle, errno, FIRMWARE_FILE_NAME, sizeof(FirmwareHeader), stdio_result );

        fclose( handle );
        return ERROR_CARD_CANNOT_READ_HDR;
    }

    // we don't need to byte swap the values
    // because host (Pentium) and card (i960)
    // are both little-endian
    //
    // but if we ever do, make sure not to swap
    // again later along with rest of firmware image
    //

    // shortcut for the load address
    //
    start = memmap->ram8 + header.start_offset;
    printf( "card 0x%4.4X firmware image will load at 0x%8.8lX and start at 0x%8.8lX\n", card_handle, start, memmap->ram8+header.entry_address );

#if COPY_FIRMWARE_HEADER_TO_CARD
    // copy the header to card memory
    //
    memcpy( start, &header, sizeof(FirmwareHeader) );

    start += sizeof( FirmwareHeader );
#endif

    // read the rest of the file into card memory via a local
    // variable because it only supports 32-bit accesses
    //
    for(
        count = sizeof(FirmwareHeader),
        current = (Bit32 *) start;

        count < size;

        count += sizeof( Bit32 ),
        current++
    )
    {
        stdio_result = fread( &one_word, sizeof(Bit32), 1, handle );
        if( stdio_result != 1 )
        {
            printf( "card 0x%4.4X error %d reading file '%s' body word %ld; got %d items\n",
                card_handle, errno, FIRMWARE_FILE_NAME, count, stdio_result );

            fclose( handle );
            return ERROR_CARD_CANNOT_READ_BODY;
        }

#if SWAP_FIRMWARE
        *current = ntohl( one_word );
#else
        *current = one_word;
#endif
    }

    fclose( handle );

    static char in_buffer[ UART_BUFFER_SIZE ];
    static char out_buffer[ 20 ];
    PentiumClock initial, now, alarm;
    double timing;
    Boot_status status;

    // suck up any console output from the ROM
    //
    // should be "Mon960\r\n"
    //
    atm_uart_io( NULL, in_buffer, UART_BUFFER_SIZE, UART_DELAY_INITIAL );

    // need to prod the NINDY monitor
    // so it will accept commands from us
    //
    // should give a signon message and a command prompt
    //
    sprintf( out_buffer, "\r" );
    atm_uart_io( out_buffer, in_buffer, NULL, UART_DELAY_INITIAL );

    // if I ever re-enable this, do it with a timeout!
    //
#if 0
    // wait for the command prompt from the NINDY monitor
    //
    do
    {
        atm_uart_io( NULL, in_buffer, UART_BUFFER_SIZE, UART_DELAY_INITIAL );
    }
    while( !strstr( in_buffer, "=>" ) );
#endif

    // tell (the NINDY OS (that the card booted after reset))
    // to start the code we downloaded
    //
    sprintf( out_buffer, "go %4.4X\r", header.entry_address & 0xFFFF );
    atm_uart_io( out_buffer, in_buffer, UART_BUFFER_SIZE, UART_DELAY_INITIAL );

    // move on to the next line of screen,
    // since NINDY monitor's prompt doesn't
    //
    printf( "\n" );

    // check for a complaint from the monitor
    //
    if( strstr( in_buffer, "fault" ) )
    {
        printf( "card 0x%4.4X ROM could not start firmware; last boot status was 0x%8.8lX\n", card_handle, status );

        // give the user a chance to play with ROM monitor manually
        //
        printf( "card 0x%4.4X is letting you talk to ROM monitor; press escape when done\n", card_handle );
        atm_uart_io( "\r", NULL, 0, UART_DELAY_FAILURE );

        return ERROR_CARD_CP_FAULTED;
    }

    // wait a few seconds for the boot status
    // to say the CP (cell processor) is running
    //
    initial.set();
    now = initial;
    alarm.put_seconds( FIRM_START_DELAY );
    alarm += now;

    while( now < alarm )
    {
        status = memmap->shared.mon960.bstat;
        if( status == cp_running )
            break;

        now.set();
    }

    if( status != cp_running )
    {
        printf( "card 0x%4.4X firmware took longer than %g seconds to run after download; last boot status was 0x%8.8lX\n", card_handle, FIRM_START_DELAY, status );
        return ERROR_CARD_CP_NOT_RUNNING;
    }

    // tell the user how long this took
    //
    now -= initial;
    timing = now.get_seconds();
    printf( "card 0x%4.4X firmware took %g S to run\n", card_handle, timing );

    return 0;
}



#if HEARTBEAT_CHECKS == 0
// turn off warning about the want_increasing argument not being used
//
#pragma warn -par
#endif

// check that the heartbeat is steadily decreasing
// or increasing
//
// assumes that wraparound will not occur now,
// since we just started the puppy, and it is a
// 32-bit value, and we hope it isn't decrementing
// very fast
//
// does not make any assumptions about how much the
// heartbeat changes each time, since that isn't documented
//
int AtmCardRunning::atm_check_heartbeat( int want_increasing )
{
#if HEARTBEAT_CHECKS != 0
    PentiumClock initial, now, alarm;
    double timing;
    Bit32 prev_heartbeat = memmap->shared.queues.heartbeat;
    Bit32 heartbeat;
    unsigned count;
    const char *direction;

    if( want_increasing )
    {
        direction = "increase";
    }
    else
    {
        direction = "decrease";
    }

    now.set();

    for( count=0; count<HEARTBEAT_CHECKS; count++ )
    {
        initial = now;
        alarm.put_seconds( BEAT_CHANGE_DELAY );
        alarm += initial;

        while( now<alarm )
        {
            heartbeat = memmap->shared.queues.heartbeat;

            now.set();

            if( want_increasing )
            {
                if( heartbeat > prev_heartbeat )
                    break;
            }
            else
            {
                if( heartbeat < prev_heartbeat )
                    break;
            }

            // suck up any output it has to give
            //
            atm_uart_io( NULL, NULL, 0, UART_DELAY_HEART );

            // wait a bit to give the card time to process,
            // since we stop the card CPU every time we
            // do a slave PCI access to card
            //
            PentiumClock::wait( BEAT_CHANGE_HOLDOFF * CLOCK_RATE );
        }

        if( now >= alarm )
        {
            printf( "card 0x%4.4X heartbeat took longer than %g seconds to %s\n", card_handle, BEAT_CHANGE_DELAY, direction );
            return ERROR_CARD_HEARTBEAT_STUCK;
        }

        alarm = now;
        alarm -= initial;
        timing = alarm.get_seconds();
        printf( "card 0x%4.4X iteration %d took %.4f S to %s from 0x%8.8lX to 0x%8.8lX\n", card_handle, count, timing, direction, prev_heartbeat, heartbeat );

        prev_heartbeat = heartbeat;
    }
#endif

    return 0;
}

#if HEARTBEAT_CHECKS == 0
// turn warning about unused arguments back on if it was on before
//
#pragma warn .par
#endif



// wait for the results of a command
// previously issued to this card
//
int AtmCardRunning::atm_await_command(
    double uart_delay,
    double command_delay,
    volatile Status *status_location
)
{
    PentiumClock now, alarm;
    double timing;
    Status status2;

    // suck up any output it has to give
    //
    atm_uart_io( NULL, NULL, 0, uart_delay );

    // wait for the card to act on the command
    //
    alarm.put_seconds( command_delay );
    alarm += atm_command_start_time;
    do
    {
        now.set();
        status2 = *status_location;
    } while( status2==stat_pending && now<alarm );

    // turn card UART usage back on, so
    // we can see output of next command
    //
    memmap->shared.queues.hlogger = TRUE;

    // check the results of command
    //
    if( status2 == stat_pending )
    {
        printf( "card 0x%4.4X '%s' cmd took longer than %g seconds\n", card_handle, atm_command_name, command_delay );
        return ERROR_CARD_CMD_STUCK;
    }

    now -= atm_command_start_time;
    timing = now.get_seconds();

    // make sure card turned off pending bit,
    // didn't turn on free bit, and turned on
    // complete bit
    //
    if( (status2 & stat_pending)
    ||  (status2 & stat_free)
    || !(status2 & stat_complete) )
    {
        printf( "card 0x%4.4X '%s' cmd transitioned to invalid state %d after %g seconds\n", card_handle, atm_command_name, status2, timing );
        return ERROR_CARD_CMD_INVALID;
    }

    // make sure command did not fail
    //
    if( status2 & stat_error )
    {
        printf( "card 0x%4.4X '%s' cmd failed after %g seconds (last heartbeat was 0x%8.8lX)\n", card_handle, atm_command_name, timing, memmap->shared.queues.heartbeat );
        return ERROR_CARD_CMD_FAILED;
    }

    printf( "card 0x%4.4X '%s' command succeeded in %g seconds\n", card_handle, atm_command_name, timing );

    return 0;
}



// issue card's "clear timestamp" command, but do not check results
// so other cards in the same machine can be checked quickly
//
int AtmCardRunning::start_clearing_timestamp( void )
{
    // check that no command is pending,
    // since we only have one status address
    //
    Status status2 = *atm_command_result;
    if( (status2 & stat_pending) || !(status2 & stat_complete) )
    {
        printf( "card 0x%4.4X cannot clear timestamp; a previous '%s' command is still pending\n", card_handle, atm_command_name );
        return ERROR_CARD_PREV_CMD_PENDING;
    }

    // remember the name of this command for later error messages
    //
    atm_command_name = "clear timestamp";

    // remember that card has not acted on our command yet
    //
    *atm_command_result = stat_pending;

    // turn off card UART usage, so this card doesn't wait
    // to print before doing the command
    //
    memmap->shared.queues.hlogger = FALSE;

    // remember when we started command
    //
    atm_command_start_time.set();

    // issue the command!
    //
    atm_command->cmd.op = op_clear_timestamp;

    return 0;
}



// wait for card's "clear timestamp" command to complete
//
int AtmCardRunning::wait_for_timestamp_to_clear( void )
{
    return atm_await_command( UART_DELAY_COMMAND, TIME_CLEAR_DELAY, atm_command_result );
}



// issue card's "shutdown" command and wait for results
//
int AtmCardRunning::atm_shutdown_command( void )
{
    PentiumClock now, alarm;
    double timing;
    Status status2;
    u_int heartbeat;

    // check that no command is pending,
    // since we only have one status address
    //
    status2 = *atm_command_result;
    if( (status2 & stat_pending) || !(status2 & stat_complete) )
    {
        printf( "card 0x%4.4X cannot shutdown; a previous '%s' command is still pending\n", card_handle, atm_command_name );
        return ERROR_CARD_PREV_CMD_PENDING;
    }

    // remember the name of this command for later error messages
    //
    atm_command_name = "shutdown";

    // mark this command as pending, in case the caller tries
    // to do some other command afterwards, which would be quite
    // invalid, since we will cause the card code to EXIT!
    //
    *atm_command_result = stat_pending;

    // remember when we started command
    //
    atm_command_start_time.set();

    // issue the command!
    //
    atm_command->cmd.op = op_shutdown;

    // wait for the card to act on the command
    //
    alarm.put_seconds( SHUTDOWN_DELAY );
    alarm += atm_command_start_time;
    do
    {
        // suck up any output it has to give, so that
        // (1) user can see it, and so (2) code on card
        // does not hang trying to print and hence never
        // exit properly
        //
        atm_uart_io( NULL, NULL, 0, 0 );

        now.set();
        heartbeat = memmap->shared.queues.heartbeat;
    }
    while( heartbeat!=SAFE_SHUTDOWN_BY_USER && now<alarm );

    // did we timeout?
    //
    if( heartbeat != SAFE_SHUTDOWN_BY_USER )
    {
        printf( "card 0x%4.4X '%s' cmd took longer than %g seconds; last heartbeat was 0x%8.8X\n", card_handle, atm_command_name, SHUTDOWN_DELAY, heartbeat );
        return ERROR_CARD_CMD_STUCK;
    }

    // suck up any final output from firmware and MON960
    //
    atm_uart_io( NULL, NULL, 0, UART_DELAY_SHUTDOWN );

    now -= atm_command_start_time;
    timing = now.get_seconds();

    // prefix this output with a newline because we MON960
    // command prompt leaves us in the middle of a line
    //
    printf( "\ncard 0x%4.4X '%s' command succeeded in %g seconds\n", card_handle, atm_command_name, timing );

    return 0;
}



// fill in parms, issue card's initialize command, check results
//
int AtmCardRunning::atm_init_command( InitType itype, LoopbackSelection ltype )
{
    Init_block *init = &memmap->shared.queues.init;

    // read the heap size (in bytes) from the card
    //
    // someday maybe we will calculate the sizes
    // that we want to use from the heap size
    //
    printf( "card 0x%4.4X heap base=0x%8.8lX size=0x%8.8lX\n",
        card_handle,
        memmap->shared.queues.heap_base,
        memmap->shared.queues.heap_size
    );

#if OLD_AALI
    // all our buffers will be in the small queue for scheme 1,
    // and they will be big enough to hold the first N cells
    // of the pcaket, which we what transferred to the host
    //
    // there probably isn't memory on the card for it to know
    // about as much of host memory at once as we would like
    // (1 MB / 1 48-byte cell = 21824 buffers would be nice)
    // so we may have to lower pool_size
    //
    init->small_b1.queue_length = 2;
    init->small_b1.buffer_size = 64;
    init->small_b1.pool_size = 60 * 2;
    init->small_b1.supply_blksize = 60;

    // the large queue for scheme 1 will be empty, so the
    // remainder of the packet will be discarded
    //
    // but we cannot set the supply block size below 4
    // and we cannot set the buffer size below ???
    //
    // I am just hoping I don't have to give it any supply blocks...
    //
    init->large_b1.queue_length = 0;
    init->large_b1.buffer_size = 0;
    init->large_b1.pool_size = 0;
    init->large_b1.supply_blksize = 0;

    // both the large and small queues for buffer scheme 2
    // will be empty all the time because we will never
    // specify that buffer scheme when activating a VP/VC
    //
    init->small_b2.queue_length = 0;
    init->small_b2.buffer_size = 0;
    init->small_b2.pool_size = 0;
    init->small_b2.supply_blksize = 0;

    init->large_b2.queue_length = 0;
    init->large_b2.buffer_size = 0;
    init->large_b2.pool_size = 0;
    init->large_b2.supply_blksize = 0;

    // we will not be using any connectionless VPI/VCI combos
    //
    init->conless_vpvc = 0;

    // we won't be transmitting so no need to extend TPD
    //
    init->tpd_extension = 0;

    // the default RPD (RX packet descriptor) holds
    // 3 RBD's (RX buffer descriptors) and since we
    // only expect one buffer to be filled per packet,
    // no extension of the RPD is appropriate
    //
    init->rpd_extension = 0;

    // allow a 2 BufferBlock's to be pending host processing
    //
    init->rqueue_len = 60 * 2;

    // we won't be doing transmit, so transmit-done
    // queue can be zero-length
    //
    init->tqueue_len = 0;

    // lets try something modest for now...
    //
    // allow 1024 receive state table entries
    //
    init->num_connect = 1024;

    // idunno, probably just for connectionless VP/VC
    // which we do not use
    //
    init->receive_threshold = 0;
#else
    // tell the card whether to enable software expansion of
    // timestamps, by incrementing a 32-bit counter when
    // the 16-bit hardware counter wraps, by enabling it to
    // interrupt the on-card i960 microprocessor
    //
    if( itype == NormalAtmWith16BitTimestamps )
        init->enable_timer_interrupt = FALSE;
    else
        init->enable_timer_interrupt = TRUE;

    // tell the card whether to loop received data
    // back out to its transmit port
    //
    if( ltype == Facility )
        init->enable_line_loopback = TRUE;
    else
        init->enable_line_loopback = FALSE;

#if CELLS_PER_PACKET == 1
    // tell the card to use 20 bits of VPI/VCI for indexing
    // when looking up AAL5 receive state
    //
    // should require 128 K bytes on card
    //
    init->index_bits_avail = 20;

    // 12 least significant of the index bits should come
    // from the VCI, leaving 8 most significant to come
    // from the least significant bits of VPI
    //
    // so we get 4096 VC's in each of 256 VP's
    //
    init->bits_from_vci = 12;
#else
    // tell the card to use 19 bits of VPI/VCI for indexing
    // when looking up AAL5 receive state
    //
    // should require 128 K bytes on card
    //
    init->index_bits_avail = 19;

    // 11 least significant of the index bits should come
    // from the VCI, leaving 8 most significant to come
    // from the least significant bits of VPI
    //
    // so we get 2048 VC's in each of 256 VP's
    //
    init->bits_from_vci = 11;
#endif

    // most significant bits of VPI and VCI which are not
    // used to index into RX state table should be zero
    //
    init->high_vpi_vci = 0;

    // tell the card about size of our buffer blocks
    //
    // location has already been filled in by atm_rx_completed()
    //
    init->host_block_size = COMPL_RING_ENTRIES * BUFFER_SIZE;
#endif  // OLD_AALI

    // we aren't going to buffer up more than one command
    // to the card, so we might as well tell the card to
    // only allocate one command queue entry, which lets
    // us not bother cycling through the entries in it,
    // so we will always use entry 0
    //
    init->cqueue_len = 1;

    // allow the card to printf() its complaints to software UART
    // (firmware images that came on floppies from FORE do not
    // seem to have such messages compiled in)
    //
    memmap->shared.queues.hlogger = TRUE;

    // remember the name of this command for later error messages
    //
    atm_command_name = "initialize";

    // remember that card has not acted on our command yet
    //
    init->status = stat_pending;

    // remember when we started command
    //
    atm_command_start_time.set();

    // issue the command!
    //
    init->op = op_initialize;

    return atm_await_command( UART_DELAY_FINAL, FIRM_INIT_DELAY, &init->status );
}



// initialize the ATM SAR chip
//
int AtmCardRunning::atm_init( InitType itype, LoopbackSelection ltype )
{
    Bit32 pointer;
    unsigned count;
    int result;
    PentiumClock initial, now, alarm;
    double timing;

#if ATM_SHIFT_CLEANUP_DEBUG
    // atm_rx_completed() has never seen buffer count overflow
    //
    bufcount_overflow_first_used = 0;
    bufcount_overflow_last_used = 0;
    bufcount_overflow_highest_used = 0;
    bufcount_nonoverflow_last_used = 0;
    bufcount_overflow_sar_rx_free = -1;
    bufcount_overflow_prev_free = -1;
    bufcount_overflow_this_free = -1;
    bufcount_overflow_next_free = -1;
    bufcount_overflow_phys_free_base = -1;
#endif

    // we haven't had any interrupts from
    // this chip on this card yet
    //
    atm_interrupt_count = 0;
    atm_interrupt_unexpected = 0;

    // we haven't had any errors from this card yet
    //
    atm_interrupt_error = 0;

    // PBI has never run out of in FIFO or out FIFO entries
    //
    atm_out_fifo_full = 0;
    atm_in_fifo_almost_full = 0;

    // show current values of some PCI configuration registers
    //
    Bit8 value;

    value = pci_get_min_grant( card_handle );
    printf( "card 0x%4.4X minimum-grant register was %d (%g uS)\n", card_handle, value, value*0.25 );

    value = pci_get_max_latency( card_handle );
    printf( "card 0x%4.4X maximum-latency register was %d (%g uS)\n", card_handle, value, value*0.25 );

    value = pci_get_max_slave_latency( card_handle );
    printf( "card 0x%4.4X maximum slave latency register was %d cycles\n", card_handle, value );

    old_pci_latency_timer = pci_get_master_latency_timer( card_handle );
    printf( "card 0x%4.4X master-latency-timer register was %d (%g uS)\n", card_handle, old_pci_latency_timer, old_pci_latency_timer*0.25 );

    value = pci_get_cache_line_size( card_handle );
    printf( "card 0x%4.4X cache line size was %d words\n", card_handle, value );

    value = pci_get_threshold( card_handle );
    printf( "card 0x%4.4X large burst or continuous request threshold was %d 32-bit FIFO words\n", card_handle, value );

    old_pci_master_control_alias = pci_get_master_control( card_handle );
    printf( "card 0x%4.4X PCI master control register was 0x%2.2X (cache line reads %s,\n"
        " write-and-invalidate %s, write-and-invalidate needs >%d lines,\n"
        " latency timer %s, continuous request mode %s,\n"
        " large PCI bursts %s forced, slave RAM accesses %s byteswapped)\n",
        card_handle,
        old_pci_master_control_alias,
        old_pci_master_control.disable_cache_line_reads   ? "disabled" : "enabled",
        old_pci_master_control.disable_write_and_invals   ? "disabled" : "enabled",
        old_pci_master_control.write_inval_needs_2_lines  ? 2          : 1,
        old_pci_master_control.ignore_latency_timer       ? "ignored"  : "respected",
        old_pci_master_control.enable_cont_request_mode   ? "enabled"  : "disabled",
        old_pci_master_control.force_large_pci_bus_bursts ? "are"      : "are not",
        old_pci_master_control.byteswap_slave_ram_access  ? "are"      : "are not"
    );

#if ATM_INIT_SETS_PCI_CONTROL
    // try to change PCI master control
    // to desired value, and check that it worked
    //
    union { PciMasterControl new_pci_master_control; Bit8 new_pci_master_control_alias; };
    new_pci_master_control_alias = old_pci_master_control_alias;
    new_pci_master_control.enable_cont_request_mode = TRUE;
    pci_set_master_control( card_handle, new_pci_master_control_alias );
    value = pci_get_master_control( card_handle );
    if( value != new_pci_master_control_alias )
        printf( "card 0x%4.4X unable to set master control register to 0x%.2X; is now 0x%.2X\n", card_handle, new_pci_master_control, value );
    else
        printf( "card 0x%4.4X enabled continuous request mode in master control register\n", card_handle );
#endif

#if ATM_INIT_SETS_PCI_TIMING
    // try to change PCI minimum-grant and maximum-latency
    // to desired values, and check that it worked
    //
    pci_set_min_grant( card_handle, DESIRED_MIN_GRANT );
    value = pci_get_min_grant( card_handle );
    if( value != DESIRED_MIN_GRANT )
        printf( "card 0x%4.4X unable to set minimum-grant register to %d; is now %d (%g uS)\n", card_handle, DESIRED_MIN_GRANT, value, value*0.25 );
    else
        printf( "card 0x%4.4X changed minimum-grant register to %d (%g uS)\n", card_handle, DESIRED_MIN_GRANT, value*0.25 );

    pci_set_master_latency_timer( card_handle, DESIRED_CARD_LATENCY );
    value = pci_get_master_latency_timer( card_handle );
    if( value != DESIRED_CARD_LATENCY )
        printf( "card 0x%4.4X unable to set master-latency-timer register to %d; is now %d (%g uS)\n", card_handle, DESIRED_CARD_LATENCY, value, value*0.25 );
    else
        printf( "card 0x%4.4X changed master-latency-timer register to %d (%g uS)\n", card_handle, DESIRED_CARD_LATENCY, value*0.25 );
#endif

    // debug
    printf( "card 0x%4.4X initial boot status at log 0x%8.8lX is 0x%8.8lX\n",
        card_handle,
        &memmap->shared.mon960.bstat,
        memmap->shared.mon960.bstat
    );
    printf( "card 0x%4.4X HCR at log 0x%8.8lX has value 0x%2.2X byte, 0x%8.8lX word\n",
        card_handle,
        &memmap->dword_00h,
        *(Bit8 *)&memmap->dword_00h,
        memmap->dword_00h
    );

#if ATM_INIT_RESETS_SAR
    // force a hardware reset
    //
    result = atm_reset_hardware();
    if( result )
        return result;

    // download firmware to card and start it
    //
    result = atm_download();
    if( result )
        return result;

    // tell the user where we think heartbeat is
    //
    printf( "card 0x%4.4X heartbeat is at logical address 0x%8.8lX\n", card_handle, &memmap->shared.queues.heartbeat );

    // make sure that the heartbeat is decreasing
    //
    result = atm_check_heartbeat( FALSE );
    if( result )
        return result;

#if CARD_INIT_PROVIDES_BUFFERS
    if( rx_allowed )
    {
        result = atm_rx_completed( during_initialization );
        if( result )
            return result;
    }
#endif

    // remember when we enabled receive
    //
    atm_enabled.set();

    // tell the card what sizes to make things,
    // and start it reading from all its queues
    //
    result = atm_init_command( itype, ltype );
    if( result )
        return result;
#endif      // ATM_INIT_RESETS_SAR

    // make sure that the heartbeat is increasing
    //
    result = atm_check_heartbeat( TRUE );
    if( result )
        return result;

    // enable per-block interrupts
    //
    // when we get this, card will already have started using
    // the second block we gave it earlier, so handler has plenty
    // of time to give it another block, so allocation can occur
    // at the beginning of interrupt handler
    //
    union { HostInterruptMask himr; Bit32 himr_alias; };
    himr_alias = 0;
#if ALLOW_ATM_RX_FREE_EMPTY_IRQ
    himr.is_masked = FALSE;
#else
    himr.is_masked = TRUE;
#endif
    memmap->dword_01h = himr_alias;

    // this old method of masking interrupts has been superceded
    // by the Host Interrupt Mask register (HIMR) at the 1MB line
    // in memory space, but we will do both in case of stupid card
    // code
    //
#if ALLOW_ATM_RX_FREE_EMPTY_IRQ
    memmap->shared.queues.imask = 1;        // 1=enabled
#else
    memmap->shared.queues.imask = 0;        // 0=disabled
#endif

    // card does not need to wait, the next
    // time it wants to interrupt us
    //
    memmap->shared.queues.istat = 0;

    // allow timer interrupts
    //
    timer_interrupts_allowed = TRUE;

    // allocate the command result in locked memory
    //
    atm_command_result = (Status *) flat_alloc( 4 );

    // last command we issued succeeded
    //
    *atm_command_result = stat_complete;

    // in case somebody is stupid enough to try printing
    // the previous command name, we better initialize it
    //
    atm_command_name = "none ever issued";

    // remember where the command queue is on the card
    //
    atm_command = (Qcard_cmd *) (memmap->ram8 + (Bit32)memmap->shared.queues.command_queue);

    // tell the card where the (only) command status is
    //
    atm_command->stataddr = flat_to_physical( atm_command_result );

    return 0;
}



void AtmCardRunning::atm_exit( void )
{
    int result;
    unsigned count;
    double timing;
    Bit8 value;
    Bit32 buffer_count, block_count;
    CardClock first, last;

#if ATM_EXIT_ITERATES_TIMER_HANDLER
    // call timer_handler() many times, as if we were the
    // interrupt handler
    //
    // warning: would cause totally inaccurate information
    // to be reported if BufferBlock didn't contain a field
    // to say how many buffers were actually filled!!!
    //
    for( count=0; count<1; count++ )
    {
        timer_handler();
    }
#endif

    // ignore timer interrupts from now on
    //
    timer_interrupts_allowed = FALSE;

#if CHECK_MEMORY_WHILE_EXITING
    // not too nice to check here since other cards
    // may still be running
    //
    all_buffers->check();
#endif

    // maybe suck the (SONET framer) statistics out of the card???
    //

#if ATM_EXIT_FILLS_PARTIAL_BLOCK
    // count any cells left over in pending block, which were not
    // quite enough to cause an pending block full interrupt
    //
    // must do this before zeroing next_host_block pointer on card!
    //
    set_pending_block_usage();
#endif

    // disable all future card interrupts
    //
    union { HostInterruptMask himr; Bit32 himr_alias; };
    himr_alias = 0;
    himr.is_masked = TRUE;
    memmap->dword_01h = himr_alias;

    // stop the card from reading from or writing to host memory
    //
    if( atm_command )
    {
        result = atm_shutdown_command();
        atm_command->stataddr = 0;
    }

#if ATM_INIT_SETS_PCI_CONTROL
    // reset PCI master control for this card
    //
    pci_set_master_control( card_handle, old_pci_master_control_alias );
    value = pci_get_master_control( card_handle );
    if( value != old_pci_master_control_alias )
        printf( "card 0x%4.4X unable to restore master control register to 0x%.2X; is now 0x%.2Xn", card_handle, old_pci_master_control_alias, value );
    else
        printf( "card 0x%4.4X restored master control register to 0x%.2X\n", card_handle, old_pci_master_control_alias );
#endif

#if ATM_INIT_SETS_PCI_TIMING
    // reset the master latency timer for this card
    //
    pci_set_master_latency_timer( card_handle, old_pci_latency_timer );
    value = pci_get_master_latency_timer( card_handle );
    if( value != old_pci_latency_timer )
        printf( "card 0x%4.4X unable to restore master-latency-timer register to %d; is now %d (%g uS)\n", card_handle, old_pci_latency_timer, value, value*0.25 );
    else
        printf( "card 0x%4.4X restored master-latency-timer register to %d (%g uS)\n", card_handle, old_pci_latency_timer, value*0.25 );
#endif





    // dump our meager statistics
    //

    if( atm_out_fifo_full )
        printf( "card 0x%4.4X PBI out FIFO was full %d times during run\n", card_handle, atm_out_fifo_full );
    else
        printf( "card 0x%4.4X PBI out FIFO was available all during run\n", card_handle );

    if( atm_in_fifo_almost_full )
        printf( "card 0x%4.4X PBI in  FIFO was almost full %d times during run\n", card_handle, atm_in_fifo_almost_full );
    else
        printf( "card 0x%4.4X PBI in  FIFO was available all during run\n", card_handle );

    printf( "card 0x%4.4X ATM last interrupt status was 0x%8.8lX\n", card_handle, atm_status_alias );
    atm_status_alias = memmap->dword_00h;
    printf( "card 0x%4.4X ATM final status 0x%8.8lX\n", card_handle, atm_status_alias );

    union { PciStatus status; Bit8 status_alias; };
    status_alias = pci_get_status( card_handle );
    printf( "card 0x%4.4X PCI final status 0x%4.4X\n", card_handle, status_alias );

    // dump the ready-to-send buffer block
    //
    printf( "card 0x%4.4X next transmit buffer block is 0x%p\n", card_handle, next_tx_block );

    // dump the currently-being-sent buffer block
    //
    printf( "card 0x%4.4X was transmitting buffer block 0x%p\n", card_handle, my_tx_block );

    // dump the ready-to-fill buffer block
    //
    printf( "card 0x%4.4X next buffer block is 0x%p\n", card_handle, next_block );

    // dump the currently-being-filled buffer block
    //
    printf( "card 0x%4.4X was filling buffer block 0x%p\n", card_handle, my_block );

#if ATM_EXIT_FILLS_PARTIAL_BLOCK
#if ATM_EXIT_DUMPS_ATM_HEADERS
    // dump the ATM headers in the completion ring
    //
    AtmHeaderCoerce *atm = (AtmHeaderCoerce *) (my_block->buffers + 8);
    for( count=0; count<my_block->used; count++ )
    {
        // if the SAR ever filled this entry in...
        //
        // we could check timestamp for nonzero
        // but that would require flows_per_block
        // to have zeroed it earlier, which is too
        // much overhead
        //
        if( TRUE )
        {
            printf( "card 0x%4.4X ATM header %d GFC=%1.1X CLP=%d PTI210=%d%d%d VCI=%d VPI=%d\n",
                card_handle,
                count,
                atm->bits.GFC,
                atm->bits.CLP,
                atm->bits.PTI2, atm->bits.PTI1, atm->bits.PTI0,
                atm->bits.VCIL + (atm->bits.VCIH<<12),
                atm->bits.VPI
            );
        }

        atm = (AtmHeaderCoerce *) (BUFFER_SIZE + (Bit8 *)atm);
    }
#endif

    // if there is a non-empty block left over...
    //
    if( my_block && my_block->used>0 )
    {
        // place my_block on global filled list
        // even though it is not quite all filled
        //
        result = all_buffers->fill( my_block );
        if( !result )
            my_block = NULL;
    }
#endif

#if ATM_EXIT_RELEASES_BLOCKS
    if( my_tx_block )
    {
        my_tx_block->state = is_consuming;      // force free to succeed
        all_buffers->free( my_tx_block );
        my_block = NULL;
    }

    if( next_tx_block )
    {
        next_tx_block->state = is_consuming;    // force free to succeed
        all_buffers->free( next_tx_block );
        next_tx_block = NULL;
    }

    if( my_block )
    {
        my_block->state = is_consuming;     // force free to succeed
        all_buffers->free( my_block );
        my_block = NULL;
    }

    if( next_block )
    {
        next_block->state = is_consuming;   // force free to succeed
        all_buffers->free( next_block );
        next_block = NULL;
    }
#endif

    printf( "card 0x%4.4X had %ld SAR interrupts, %ld of which were unexpected\n",
        card_handle, atm_interrupt_count, atm_interrupt_unexpected );
//    if( atm_interrupt_count > 0 )
    {
        printf( "card 0x%4.4X last error during interrupt was %d\n", card_handle, atm_interrupt_error );

    #if ESSENTIAL_TIME_STAMPS
        double elapsed_min2, elapsed_max2, elapsed_avg2;
        printf( "card 0x%4.4X had %ld timer interrupts\n", card_handle, timer_interrupt_count );
        if( timer_interrupt_count )
        {
            elapsed_min2 = ((double)elapsed_min * 1000) / CLOCK_RATE;
            elapsed_max2 = ((double)elapsed_max * 1000) / CLOCK_RATE;
            elapsed_avg2 = (((double)elapsed_tot / timer_interrupt_count) * 1000) / CLOCK_RATE;
            printf( "card 0x%4.4X timer_handler() took min %.4f max %.4f avg %.4f mS\n", card_handle, elapsed_min2, elapsed_max2, elapsed_avg2 );
        }
    #else
        printf( "card 0x%4.4X sorry, AtmCardRunning::handler() timing is unavailable\n", card_handle );
    #endif
    }

    // dump the host buffer queue states
    //
    buffer_count = all_buffers->amount_filled( card_handle, block_count );
    printf( "card 0x%4.4X has %ld/%ld filled buffers/blocks in host queue\n", card_handle, buffer_count, block_count );

#if CARD_KEEPS_FIFO_STATS
    int seen = 0;
    for( count=0; count<=MAX_PHY_FIFO_DEPTH; count++ )
        if( memmap->shared.queues.fifo_depth_counts[ count ] > 0 )
        {
#if SHOW_FIFO_DEPTHS
            printf( "card 0x%.4X counted FIFO depth %d %d times\n", card_handle, count, memmap->shared.queues.fifo_depth_counts[count] );
#endif
            seen++;
        }
    printf( "card 0x%.4X used %d FIFO depths\n", card_handle, seen );
#endif

    BufferBlock *calibration_block = all_buffers->read( card_handle, ATM_CALIBRATE - 1 );
    if( !calibration_block )
        printf( "card 0x%4.4X did not fill %d buffer blocks, so calibration cannot be done\n", card_handle, ATM_CALIBRATE );
    else
    {
        last = calibration_block->last_timestamp();
        first = calibration_block->first_timestamp();
        last.strip_bitfields();
        first.strip_bitfields();
        timing = CardClock( last - first ).get_seconds();
        printf( "card 0x%4.4X took %g S total between filling the first and %dth buffer in %dth block\n", card_handle, timing, calibration_block->used, ATM_CALIBRATE );

        if( calibration_block->used != 0 )
        {
            timing /= calibration_block->used;
            printf( "card 0x%4.4X took %g S to fill each buffer in that one block\n", card_handle, timing );
            timing = 1 / timing;
            printf( "card 0x%4.4X saw %g packets per S in block %d\n", card_handle, timing, ATM_CALIBRATE );
        }
#if ATM_CALIBRATE>=2
        timing = CardClock( calibration_block->last_timestamp()
               - all_buffers->read( card_handle, (unsigned)0 )->first_timestamp() ).get_seconds();
        printf( "card 0x%4.4X took %g S total between filling the first buffer of first block and %dth buffers of %dth block\n", card_handle, timing, calibration_block->used, ATM_CALIBRATE );
        timing /= ((ATM_CALIBRATE - 1) * (double)COMPL_RING_ENTRIES) + calibration_block->used;
        printf( "card 0x%4.4X took %g S to fill each buffer\n", card_handle, timing );
        timing = 1 / timing;
        printf( "card 0x%4.4X saw %g packets per S in first %d blocks\n", card_handle, timing, ATM_CALIBRATE );
#endif
    }






    // count fifo depths across all blocks
    //
    Bit16 fifo_depth;
    Bit32 fifo_depth_counts[ MAX_FIFO_DEPTH + 1 ];
    for( count=0; count<=MAX_FIFO_DEPTH; count++ )
        fifo_depth_counts[ count ] = 0;

    Bit32 sum_used = 0;

    CardClock prev_when, this_when, next_when;
    CardClock delta_when, max_delta, min_delta, sum_delta;
    min_delta.low  = BIT32_MAX;
    min_delta.high = BIT32_MAX;
    max_delta.low  = 0;
    max_delta.high = 0;
    sum_delta.low  = 0;
    sum_delta.high = 0;

    for( block_count=0; block_count<SW_RINGS; block_count++ )
    {
        BufferBlock *calibration_block = all_buffers->read( card_handle, block_count );
        if( !calibration_block )
            break;

        sum_used += calibration_block->used;

        tracerec *data = (tracerec *) calibration_block->buffers;

        if( block_count == 0 )
        {
            prev_when = data[0].when;
            prev_when.strip_bitfields();
            count = 1;
        }
        else
        {
            count = 0;
        }
        this_when = data[count].when;
        fifo_depth = this_when.strip_bitfields_gimme();

        for( ; count<calibration_block->used; count++ )
        {
#if !CARD_HANDLES_TIMESTAMP_WRAP
            if( this_when.low < prev_when.low )
            {
                // assumes hardware timestamp is 16 bits
                // and the field it is in is >16 bits
                //
                // note that this does not fix the absolute
                // timestamp of current record, just its value
                // relative to previous record
                //
                prev_when.low -= 65536;
                prev_when.high -= 1;
            }
#endif

            delta_when = CardClock( this_when - prev_when );

#if ATM_EXIT_DUMPS_TIMES
            if( fifo_depth > 1 )
            {
                if( kbhit() )
                    break;

                printf( "card 0x%.4X noticed FIFO depth %d at timestamp 0x%X%.8X delta %d\n",
                    card_handle,
                    fifo_depth,
                    this_when.high, this_when.low,
                    delta_when.low
                );
            }
#endif
            if( fifo_depth > MAX_FIFO_DEPTH )
                fifo_depth = MAX_FIFO_DEPTH;
            fifo_depth_counts[ fifo_depth ]++;
            next_when = data[count+1].when;
            fifo_depth = next_when.strip_bitfields_gimme();

#if CARD_HANDLES_TIMESTAMP_WRAP
            if( this_when < prev_when )
            {
                printf( "card 0x%.4X ignoring out-of-order timestamp 0x%X%.8X in record %d of block %d;"
                    "prev/next timestamps were 0x%X%.8X/0x%X%.8X\n",
                    card_handle,
                    this_when.high, this_when.low,
                    count, block_count,
                    prev_when.high, prev_when.low,
                    next_when.high, next_when.low
                );
         //       continue;
            }
#endif

            sum_delta += delta_when;
            if( delta_when > max_delta )
                max_delta = delta_when;
            if( delta_when < min_delta )
                min_delta = delta_when;
            prev_when = this_when;
            this_when = next_when;
        }
#if ATM_EXIT_DUMPS_TIMES
        if( kbhit() )
        {
            printf( "card 0x%.4X stopped examining %dth block because of keypress\n", card_handle, block_count );
            getch();
        }
#endif

    }   // end for all blocks

    if( sum_used > 0 )
    {
        timing = sum_delta.get_seconds() / sum_used;
        printf( "card 0x%4.4X timestamp delta in all blocks was %g/%g/%g S min/max/avg\n",
            card_handle,
            min_delta.get_seconds(), max_delta.get_seconds(), timing
        );

#if SHOW_FIFO_DEPTHS
        for( count=0; count<=MAX_FIFO_DEPTH; count++ )
            if( fifo_depth_counts[ count ] > 0 )
                printf( "card 0x%.4X host counted FIFO depth %d %d times across all blocks\n", card_handle, count, fifo_depth_counts[count] );
#endif
    }
    else
    {
        printf( "card 0x%.4X captured no data, so we cannot calculate"
            " timestamp deltas or FIFO depth counts\n", card_handle );
    }






    // dump some of the queue values in Cp_queue???
    //

#if SAVE_INTERRUPT_THIS_PTR
    // dump the "this" pointer seen by the interrupt handler
    //
    // if it doesn't match up with the value printed by pci_init()
    // we have trouble in river city!
    //
    printf( "'this' pointer last seen by interrupt handler was 0x%p\n", atm_this_pointer_during_interrupt );
    printf( "card 0x%4.4X correct 'this' pointer is 0x%p\n", card_handle, this );
#endif

#if ATM_SHIFT_CLEANUP_DEBUG
    if( bufcount_nonoverflow_last_used != 0 )
    {
        timing = (double) (bufcount_nonoverflow_last_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count last nonoverflow to %d %g mS after enabling RX\n", card_handle, bufcount_nonoverflow_last_used, timing );
    }

    if( bufcount_overflow_first_used != 0 )
    {
        timing = (double) (bufcount_overflow_first_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count first overflowed to %d %g mS after enabling RX\n", card_handle, bufcount_overflow_first_used, timing );
    }

    if( bufcount_overflow_last_used != 0 )
    {
        timing = (double) (bufcount_overflow_last_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count last overflowed to %d %g mS after enabling RX\n", card_handle, bufcount_overflow_last_used, timing );
    }

    if( bufcount_overflow_highest_used != 0 )
    {
        timing = (double) (bufcount_overflow_highest_time - atm_enabled) / (CLOCK_RATE / 1000.0);
        printf( "card 0x%4.4X buffer count highest overflowed to %d %g mS after enabling RX\n", card_handle, bufcount_overflow_highest_used, timing );
        printf( "card 0x%4.4X buffer count overflowed when SAR RX free ptr was 0x%8.8lX\n", card_handle, bufcount_overflow_sar_rx_free );
        printf( "card 0x%4.4X buffer count overflowed when prev_free was %d\n", card_handle, bufcount_overflow_prev_free );
        printf( "card 0x%4.4X buffer count overflowed when this_free was %d\n", card_handle, bufcount_overflow_this_free );
        printf( "card 0x%4.4X buffer count overflowed when next_free was %d\n", card_handle, bufcount_overflow_next_free );
        printf( "card 0x%4.4X buffer count overflowed when my_free_base was phys 0x%8.8lX\n", card_handle, bufcount_overflow_phys_free_base );
    }
#endif

#if DEBUG_STACK_OVERFLOW
    printf( "stack pointer inside first AtmCardRunning::timer_handler() was 0x%4.4X:%8.8lX\n", stack_seg_during_class_timer, stack_ptr_during_class_timer );
#endif
}



void AtmCardRunning::atm_handler( void )
{
    int result;

    // remember CPU time when clock interrupted us as early
    // in the handler as possible, so we have the best shot
    // at card clock calibration
    //
    // remember we cannot access stack variables at interrupt
    // time because the compiler assumes stack segment register
    // is same as data segment, and it isn't
    //
    before.set();

#if SAVE_INTERRUPT_THIS_PTR
    atm_this_pointer_during_interrupt = this;
#endif

#if ESSENTIAL_STATUS_QUERIES
    // get the status so the foreground can tell
    // whether RX overrun occurred before last cell
    //
    // probably only useful for code debug
    //
    // definitely slows down card by stopping its i960
    // CPU to do the slave PCI read
    //
    atm_status_alias = memmap->dword_00h;

    if( atm_status.out_fifo_full )
        atm_out_fifo_full++;

    if( atm_status.in_fifo_almost_full )
        atm_in_fifo_almost_full++;
#endif

    // if the card needs next rx block filled in...
    //
    if( rx_allowed && (!tx_allowed || !memmap->shared.queues.next_host_block) )
    {
        result = atm_rx_completed( during_interrupt );
        if( result && !atm_interrupt_error )
            atm_interrupt_error = result;

        // if the card is working on a block now...
        //
        if( my_block )
        {
            // get the timestamp of the packet that caused
            // the card to switch blocks, which is now the
            // first in the new block
            //
            // remember that on a slow link, the packet before
            // this one might have arrived 10 minutes ago!
            //
            atm_interrupt_time = * (CardClock *) my_block->buffers;
            atm_interrupt_time.strip_bitfields();

            // compensate for card clock rate drift
            //
            atm_interrupt_time.recalibrate_clock_rate( before );
        }
    }

    // if the card needs next tx block filled in
    // and transmit is enabled...
    //
    if( tx_allowed && (!rx_allowed || !memmap->shared.queues.next_tx_block) )
    {
        result = atm_tx_completed( during_interrupt );
        if( result && !atm_interrupt_error )
            atm_interrupt_error = result;
    }

    // acknowledge the interrupt
    //
    // machine resets if we don't do this!
    //
    union { HostControl_write hcr_w; Bit32 hcr_w_alias; };
    hcr_w_alias = 0;
    hcr_w.clear_host_int = TRUE;
    memmap->dword_00h = hcr_w_alias;

    // count how many times this occurred
    //
    atm_interrupt_count++;
}



// called by foreground code when it notices that
// BufferQueue::read_tx() has failed and it has
// just filled up a buffer
//
// note that only the card whose buffer got filled
// gets called, since there are separate tx lists
// for each card, unlike the filled list which is
// integrated
//
//
//
// also called by foreground code to enable transmit
// for the first time if we didn't give the card tx
// buffers before issuing the initialization command
//
void AtmCardRunning::foreground_tx_replenish( void )
{
    int result;

    // if this card will interrupt us again...
    //
    if( memmap->shared.queues.next_tx_block )

        // no need for any extra effort on our part
        //
        return;

    // try to get a BufferBlock
    //
    // when the card sees the next cell, it will grab
    // the block and interrupt us to give it another one
    //
    // note that this will cause another buffer read_tx
    // if there isn't really one available, but that is OK
    // since the caller needs the flag set again in order
    // to know to call us again
    //
    result = atm_tx_completed( during_interrupt );
    if( result && !atm_interrupt_error )
        atm_interrupt_error = result;
}



// called by foreground code when it notices that
// BufferQueue::allocate() has failed and it has
// just freed up a buffer
//
// note that both cards are called even when only
// one buffer has been freed up
//
//
//
// also called by foreground code to enable receive
// for the first time if we didn't give the card buffers
// before issuing the initialization command
//
void AtmCardRunning::foreground_replenish( void )
{
    int result;

    // if this card will interrupt us again...
    //
    if( memmap->shared.queues.next_host_block )

        // no need for any extra effort on our part
        //
        return;

    // try to get a BufferBlock
    //
    // when the card sees the next cell, it will grab
    // the block and interrupt us to give it another one
    //
    // note that this will cause another buffer allocation
    // if there isn't really one available, but that is OK
    // since the caller needs the flag set again in order
    // to know to call us again
    //
    result = atm_rx_completed( during_interrupt );
    if( result && !atm_interrupt_error )
        atm_interrupt_error = result;
}



// manages 2 BufferBlock's each of which has 17K buffers;
// both have been given to the card, and in the steady-state
// my_tx_block is being transmitted, and next_tx_block is on deck
//
// however, the edge condition, where the card has completed
// sending my_tx_block and started next_tx_block, is when we get
// a host interrupt and this code gets called
//
//
//
// this code does not deal with buffer exhaustion, so if the
// host ever runs out of BufferBlock's, the card will too!
//
// it is also unsafe to call this code when the card has NOT
// swapped its pointers, since
//
int AtmCardRunning::atm_tx_completed( AtmShiftTiming timing )
{
    int result, next_result;

    if( timing == during_initialization )
    {
        result = all_buffers->read_tx( &my_tx_block, card_index );
        if( result )
            return result;

        my_tx_block->cardnum = card_handle;

        memmap->shared.queues.current_tx_block = flat_to_physical( my_tx_block->buffers );
        memmap->shared.queues.current_tx_length = my_tx_block->used * WORDS_PER_SEND;
    }
    else
    {
        // if the card has any buffers...
        //
        if( my_tx_block )
        {
            // assume all of the buffers were sent
            //
            my_tx_block->freed = SEND_RING_ENTRIES;

            // place my_tx_block on global free list
            //
            result = all_buffers->free( my_tx_block );
            if( result != 0 )
                return result;
        }

        // shift the block pointers
        //
        my_tx_block = next_tx_block;
    }

    // read another next_tx_block
    //
    next_result = all_buffers->read_tx( &next_tx_block, card_index );
    if( !next_result )
    {
        next_tx_block->cardnum = card_handle;

        // tell the card about it
        //
        memmap->shared.queues.next_tx_block = flat_to_physical( next_tx_block->buffers );
        memmap->shared.queues.next_tx_length = next_tx_block->used * WORDS_PER_SEND;
    }

    return next_result;
}



// manages 2 BufferBlock's each of which has 17K buffers;
// both have been given to the card, and in the steady-state
// my_block is being filled in, and next_block is on deck
//
// however, the edge condition, where the card has completed
// filling my_block and started next_block, is when we get
// a host interrupt and this code gets called
//
//
//
// this code does not deal with buffer exhaustion, so if the
// host ever runs out of BufferBlock's, the card will too!
//
// it is also unsafe to call this code when the card has NOT
// swapped its pointers, since
//
int AtmCardRunning::atm_rx_completed( AtmShiftTiming timing )
{
    int result, next_result;

    if( timing == during_initialization )
    {
        result = all_buffers->allocate( &my_block );
        if( result )
            return result;

        my_block->cardnum = card_handle;
        my_block->cardindex = card_index;

        memmap->shared.queues.current_host_block = flat_to_physical( my_block->buffers );
    }
    else
    {
        // if the card has any buffers...
        //
        if( my_block )
        {
            // assume all of the buffers were filled
            //
            my_block->used = COMPL_RING_ENTRIES;

            // place my_block on global filled list
            //
            result = all_buffers->fill( my_block );
            if( result != 0 )
                return result;
        }

        // shift the block pointers
        //
        my_block = next_block;
    }

    // allocate another next_block
    //
    next_result = all_buffers->allocate( &next_block );
    if( !next_result )
    {
        // remember which card was expected to fill in the block
        //
        next_block->cardnum = card_handle;
        next_block->cardindex = card_index;

        // tell the card about it
        //
        memmap->shared.queues.next_host_block = flat_to_physical( next_block->buffers );
    }

    return next_result;
}



// no idea who would want to call this
//
// we actually have 2 choices here: we could use the
// count of buffers that had DMA queued up (by looking at
// current_tx_block or current_tx_length) or we could
// use the count of buffers that actually got sent to
// the SONET framer (by using current_tx_sent)
//
// I chose the latter
//
void AtmCardRunning::set_pending_tx_block_usage( void )
{
    if( my_tx_block )
    {
        Bit32 current = memmap->shared.queues.current_tx_sent;
        my_tx_block->freed = current / WORDS_PER_SEND;
#if PENDING_BLOCK_PARANOIA
        if( my_tx_block->freed > SEND_RING_ENTRIES )
        {
            printf( "card 0x%.4X internal error: calculated %d usage in pending RX block; truncated to max %d\n", card_handle, my_tx_block->freed, SEND_RING_ENTRIES );
            my_tx_block->freed = SEND_RING_ENTRIES;
        }
#endif
    }
}



// called when main() wants to flush flows for a
// partially-filled block because it got queried
//
void AtmCardRunning::set_pending_block_usage( void )
{
    if( my_block )
    {
        Bit32 current = memmap->shared.queues.current_host_block;
        if( !current )
            my_block->used = COMPL_RING_ENTRIES;
        else
        {
            my_block->used = (current - flat_to_physical( my_block->buffers )) / BUFFER_SIZE;
#if PENDING_BLOCK_PARANOIA
            if( my_block->used > COMPL_RING_ENTRIES )
            {
                printf( "card 0x%.4X internal error: calculated %d usage in pending RX block; truncated to max %d\n", card_handle, my_block->used, COMPL_RING_ENTRIES );
                printf( "card 0x%.4X internal error: current=0x%.8X buffers=0x%.8X phys\n", card_handle, current, flat_to_physical(my_block->buffers) );
                my_block->used = COMPL_RING_ENTRIES;
            }
#endif
        }
    }
}



// nonstatic timer interrupt handler
//
// DANGER!  accessing stack variables, including
// the hidden ones used for structure return values,
// causes exception E and locks up machine!
//
void AtmCardRunning::timer_handler( void )
{
#if DEBUG_STACK_OVERFLOW
    if( !stack_ptr_recorded )
    {
        //stack_ptr_recorded = TRUE;
        stack_ptr_during_class_timer = _ESP;
        stack_seg_during_class_timer = _SS;
    }
#endif

    if( !timer_interrupts_allowed )
        return;

#if DISABLE_TIMER_REENTRY
    timer_interrupts_allowed = FALSE;
#endif

#if ESSENTIAL_TIME_STAMPS
    before.set();
#endif

#if ESSENTIAL_STATUS_QUERIES
    // get the status so the foreground can tell
    // whether RX overrun occurred before last cell
    //
    // probably only useful for code debug
    //
    atm_status_alias = memmap->dword_00h;

    if( atm_status.out_fifo_full )
        atm_out_fifo_full++;

    if( atm_status.in_fifo_almost_full )
        atm_in_fifo_almost_full++;
#endif

    int result = atm_rx_completed( during_interrupt );
    if( result != 0 )
        atm_interrupt_error = result;

#if ESSENTIAL_TIME_STAMPS
    elapsed.set();
    timer_interrupt_count++;
    elapsed -= before;
    elapsed_tot += elapsed;
    if( elapsed > elapsed_max )
        elapsed_max = elapsed;
    if( elapsed < elapsed_min )
        elapsed_min = elapsed;
#endif

#if DISABLE_TIMER_REENTRY
    timer_interrupts_allowed = TRUE;
#endif
}



// turn off warning about the mode argument not being used
//
#pragma warn -par

int AtmCardRunning::card_init(
    InitType itype,
    LoopbackSelection ltype,
    int want_rx0,
    int want_tx0
)
{
    int result;

    rx_allowed = want_rx0;
    tx_allowed = want_tx0;

    sonet_init();
    printf( "card 0x%4.4X SONET initialized\n", card_handle );

    result = atm_init( itype, ltype );
    printf( "card 0x%4.4X ATM initialized with result %d\n", card_handle, result );

    return result;
}

// turn warning about unused arguments back on if it was on before
//
#pragma warn .par


// constructor
//
AtmCardRunning::AtmCardRunning( Bit8 card_index0 )
:
    AtmCard( card_index0 ),
    timer_interrupts_allowed( FALSE ),
    atm_uart_new_line( TRUE ),
    atm_command( NULL )
{
    // fix up the index to be zero-based,
    // now that we are attached to the correct card
    //
    card_index -= ATM_NIC_OFFSET;

    // first make sure that our base class initialized OK
    //
    if( init_result )
        return;

    // MUST initialize these before returning
    // so atm_exit() can use these,
    // so lets do it before any errors can occur
    //
    my_block = 0;
    next_block = 0;
    my_tx_block = 0;
    next_tx_block = 0;

#if ESSENTIAL_TIME_STAMPS
    elapsed_min.high = -1;
    elapsed_min.low  = -1;
    elapsed_max.high = 0;
    elapsed_max.low  = 0;
    elapsed_tot.high = 0;
    elapsed_tot.low  = 0;
    timer_interrupt_count = 0;
#endif

    interrupt_init();
}



// destructor
//
AtmCardRunning::~AtmCardRunning()
{
    // it is quite unsafe to exit if we never initialized!
    //
    if( init_result )
        return;

    // clean up ATM SAR chip
    //
    atm_exit();

    // clean up SONET framer chip
    //
    sonet_exit();

    // clean up interrupt grabbing
    //
    interrupt_exit();
}



AtmCard::AtmCard( Bit8 card_index0 )
{
    card_index = card_index0;

    // find the PCI card we want
    //
    card_handle = pci_get_device_handle( FORE_PCA200E_DEVICE_ID, FORE_VENDOR_ID, card_index );
    if( card_handle == RETCODE_NO_MORE_PCI_DEVICES )
    {
        fprintf( stderr, "cannot find vendor ID 0x%.4X, device ID 0x%.4X, index %d\n", FORE_VENDOR_ID, FORE_PCA200E_DEVICE_ID, card_index );
        init_result = ERROR_CARD_NOT_FOUND;
        return;
    }
    fprintf( stdout, "PCI BIOS handle for FORE PCA200E card %d is 0x%.4X\n", card_index, card_handle );

    // get the physical addresses of the one region we care about
    //
    phys_address = pci_get_base_address( card_handle );
    if( phys_address == 0 )
    {
        fprintf( stderr, "PCA200E is not mapped into host memory!\n" );
        init_result = ERROR_CARD_UNMAPPED;
        return;
    }
    fprintf( stdout, "PCA200E is mapped into host memory at phys 0x%.8X\n", phys_address );

    // get accessibility to the memory region
    //
    memmap = (pca200e_memmap *) map_physical( phys_address, sizeof(pca200e_memmap), &phys_handle );
    if( memmap == NULL )
    {
        fprintf( stderr, "cannot map %d bytes at phys address 0x%.8X into flat memory for PCA200E %d\n", sizeof(pca200e_memmap), phys_address, card_index );
        init_result = ERROR_CARD_CANNOT_MAP_PHYS;
        return;
    }

    init_result = 0;
}



AtmCard::~AtmCard()
{
    // it is quite unsafe to exit if we never initialized!
    //
    if( init_result )
        return;

    // give up addressibility for card
    //
    // assumes that cards are deleted in the reverse
    // of the order they were created in, because the
    // physical memory access routines require that
    //
    unmap_physical( sizeof(pca200e_memmap), phys_handle );
}



